/*
								+----------------------------------+
								|                                  |
								|  *** Basic compression algs ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2008   |
								|                                  |
								|           Compress.cpp           |
								|                                  |
								+----------------------------------+
*/

/*
 *	2007-02-25
 *
 *	this is first beta version if the file. todo - rewrite TBuffer so it can realloc
 *	itself in a way std::vector can and try to avoid any (re)allocations while (un)packing
 *
 *	2008-03-13
 *
 *	TBuffer was rewritten as requested, fixed some signed / unsigned mismatches for gcc
 *
 *	2008-11-20
 *
 *	TBuffer unit is no longer unsigned char, but uint8_t instead, this should avoid
 *	incompatibility with some extended character encoding in newer versions of visual studio
 *
 */

#include <assert.h>
#define _ASSERTE(e) assert(e)
#include <vector>
#include <algorithm>
#include <limits.h>
#include "Thread.h"
#include "Compress.h"
#include "MinMax.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif
// msvc 'for' scoping hack

/*
 *								=== TBuffer ===
 */

/*
 *	TBuffer::TBuffer()
 *		- default constructor, initializes empty buffer
 */
TBuffer::TBuffer()
	:p_data(0), n_size(0), n_capacity(0)
{}

/*
 *	TBuffer::TBuffer(uint32_t _n_size)
 *		- creates a new buffer with size _n_size bytes
 *		- note buffer is going to be empty in case there was not enough memory
 */
TBuffer::TBuffer(uint32_t _n_size)
{
	n_capacity = n_size = (p_data = new uint8_t[_n_size])? _n_size : 0;
	//printf("<a>\n");
}

/*
 *	TBuffer::TBuffer(const TBuffer &r_t_buffer)
 *		- copy-constructor
 *		- note buffer is going to be empty in case there was not enough memory
 */
TBuffer::TBuffer(const TBuffer &r_t_buffer)
{
	n_capacity = n_size = (p_data = new uint8_t[r_t_buffer.n_size])? r_t_buffer.n_size : 0;
	memcpy(p_data, r_t_buffer.p_data, n_size * sizeof(uint8_t));
	//printf("<a>\n");
}

/*
 *	TBuffer &TBuffer::operator =(const TBuffer &r_t_buffer)
 *		- copy-operator
 *		- note buffer is going to be empty in case there was not enough memory
 */
TBuffer &TBuffer::operator =(const TBuffer &r_t_buffer)
{
	if(p_data)
		delete[] p_data;
	n_capacity = n_size = (p_data = new uint8_t[r_t_buffer.n_size])? r_t_buffer.n_size : 0;
	//printf("<a>\n");
	memcpy(p_data, r_t_buffer.p_data, n_size * sizeof(uint8_t));
	return *this;
}

/*
 *	TBuffer::~TBuffer()
 *		- destructor
 *		- deletes buffer data
 */
TBuffer::~TBuffer()
{
	if(p_data)
		delete[] p_data;
}

/*
 *	void TBuffer::Swap(TBuffer &r_t_buffer)
 *		- swaps contents of this and r_t_buffer
 */
void TBuffer::Swap(TBuffer &r_t_buffer)
{
	uint32_t n_tmp_size = n_size;
	n_size = r_t_buffer.n_size;
	r_t_buffer.n_size = n_tmp_size;
	uint32_t n_tmp_capa = n_capacity;
	n_capacity = r_t_buffer.n_capacity;
	r_t_buffer.n_capacity = n_tmp_capa;
	uint8_t *p_tmp_data = p_data;
	p_data = r_t_buffer.p_data;
	r_t_buffer.p_data = p_tmp_data;
}

/*
 *	bool TBuffer::Resize(uint32_t n_new_size, bool b_data_important = true)
 *		- changes size of the buffer
 *		- note that resizing below buffer capacity never causes reallocations
 *		- if b_data_important is set, buffer data are copied when reallocating,
 *		  otherwise they're not (speed optimization)
 *		- returns true on success, false on failure (data aren't lost on failure,
 *		  but buffer size remains unchaged)
 */
bool TBuffer::Resize(uint32_t n_new_size, bool b_data_important)
{
	if(n_new_size <= n_capacity) {
		n_size = n_new_size;
		return true;
	}
	// don't have to resize

	uint32_t n_new_capacity = (n_capacity)? n_capacity : n_new_size;
	while(n_new_capacity < n_new_size) {
		if(n_new_capacity > UINT32_MAX / 2 ||
		   (n_new_capacity == UINT32_MAX / 2 && UINT32_MAX % 2)) {
			n_new_capacity = UINT32_MAX;
			break;
		} else
			n_new_capacity *= 2;
	};
	// calculate new capacity

	uint8_t *p_new_data;
	if(!(p_new_data = new uint8_t[n_new_capacity]))
		return false;
	//printf("<a>\n");
	// alloc

	if(p_data) {
		if(b_data_important)
			memcpy(p_new_data, p_data, n_size * sizeof(uint8_t));
		delete[] p_data;
	}
	// dealloc, copy data if required

	p_data = p_new_data;
	n_size = n_new_size;
	n_capacity = n_new_capacity;
	// set new buffer

	return true;
}

/*
 *	bool TBuffer::Grow(uint32_t n_add)
 *		- resizes the buffer so it contains at least n_add more free bytes
 *		- returns true on success, false on failure (data aren't lost on failure,
 *		  but buffer size remains unchaged)
 */
bool TBuffer::Grow(uint32_t n_add)
{
	if(n_size > UINT32_MAX - n_add)
		return false; // overflow
	uint32_t n_required_size = n_size + n_add;
	if(!Resize(n_required_size))
		return false;
	n_size = n_capacity;
	// resize to max

	return true;
}

/*
 *	void TBuffer::Clear()
 *		- clears contents of the buffer (both size and capacity will be 0)
 */
void TBuffer::Clear()
{
	if(p_data)
		delete[] p_data;
	p_data = 0;
	n_size = 0;
	n_capacity = 0;
	// clears the buffer
}

/*
 *	uint32_t TBuffer::n_Max_Size()
 *		- returns the greatest size possible buffer can be allocated to
 */
uint32_t TBuffer::n_Max_Size()
{
	return UINT32_MAX;
}

/*
 *								=== ~TBuffer ===
 */

/*
 *								=== CBurrowsWheelerTransform ===
 */

/*
 *	CBurrowsWheelerTransform::CWrapMemCmp
 *		- function object implementing less-than ordering
 *		  for indices, pointing to round buffer
 */
class CBurrowsWheelerTransform::CWrapMemCmp {
protected:
	const uint8_t *m_p_org, *m_p_end;
	uint32_t m_n_size;

public:
	/*
	 *	inline CWrapMemCmp::CWrapMemCmp(const TBuffer &r_t_data_buffer)
	 *		- default constructor
	 *		- we're going to sort indices pointing to r_t_data_buffer
	 */
	inline CWrapMemCmp(const TBuffer &r_t_data_buffer)
		:m_p_org(r_t_data_buffer.p_Data()),
		m_p_end(r_t_data_buffer.p_Data() + r_t_data_buffer.n_Size()),
		m_n_size(r_t_data_buffer.n_Size())
	{}

	/*
	 *	inline bool CWrapMemCmp::operator ()(uint32_t n_off_a,
	 *		uint32_t n_off_b) const
	 *		- less-than ordering operator
	 */
	inline bool operator ()(uint32_t n_off_a, uint32_t n_off_b) const
	{
		if(n_off_a == n_off_b)
			return false;
		// they are equal then

		const uint8_t *p_org = m_p_org, *p_end = m_p_end;
		// antialiass

		const uint8_t *p_byte_a = p_org + n_off_a;
		const uint8_t *p_byte_b = p_org + n_off_b;
		// get pointers

		for(uint32_t i = m_n_size; i; -- i) {
			if(*p_byte_a < *p_byte_b)
				return true;
			else if(*p_byte_a > *p_byte_b)
				return false;
			// compare

			++ p_byte_a, ++ p_byte_b;
			// increment

			if(p_byte_a == p_end)
				p_byte_a = p_org;
			else if(p_byte_b == p_end) // n_off_a != n_off_b
				p_byte_b = p_org;
			// wrap
		}

		return false;
	}
};

/*
 *	CBurrowsWheelerTransform::CIota
 *		- function object for creating ascending integer sequences
 */
class CBurrowsWheelerTransform::CIota {
protected:
	int m_n_counter;

public:
	inline CIota(int n_value)
		:m_n_counter(n_value)
	{}

	inline int operator ()()
	{
		return m_n_counter ++;
	}
};

/*
 *	static bool CBurrowsWheelerTransform::Decode(const TBuffer &r_t_in_buffer,
 *		TBuffer &r_t_out_buffer)
 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- returns true on success, false on failure
 */
bool CBurrowsWheelerTransform::Decode(const TBuffer &r_t_src, TBuffer &r_t_dest)
{
	_ASSERTE(&r_t_src != &r_t_dest);
	_ASSERTE(r_t_src.p_Data());
	// make some assumptions about source buffer

	if(r_t_src.n_Size() < sizeof(uint32_t))
		return false;
	uint32_t n_primary_index = *(uint32_t*)r_t_src.p_Data();
	// get primary index

	const uint8_t *p_src_data = r_t_src.p_Data() + sizeof(uint32_t);
	uint32_t n_src_size = r_t_src.n_Size() - sizeof(uint32_t);
	// get pointer to and size of the real data

	if(n_primary_index >= n_src_size)
		return false;
	// see if primary index is valid

	if(!r_t_dest.Resize(n_src_size, false))
		return false;
	// alloc output buffer

	const int n_max_word_value = 1 << (8 * sizeof(uint8_t));
	// number of combinations in a single word

	uint32_t p_buckets[n_max_word_value] = {0};

	std::vector<uint32_t> indices_list;
	indices_list.reserve(n_src_size);
	if(indices_list.capacity() < n_src_size)
		return false;
	// allocate list for indices

	{
		const uint8_t *p_src = p_src_data,
			*p_end = p_src_data + n_src_size;
		for(; p_src != p_end; ++ p_src) {
			indices_list.push_back(p_buckets[*p_src]);
			if(!(++ p_buckets[*p_src]))
				return false; // overflow
		}
	}
	// generate indices

	{
		uint32_t n_sum = 0;
 		for(int i = 0; i < n_max_word_value; ++ i) {
			uint32_t n_value = p_buckets[i];
			p_buckets[i] = n_sum;
			if(n_sum > UINT32_MAX - n_value)
				return false; // overflow
			n_sum += n_value;
		}
	}
	// integrate buckets

	{
		const uint8_t *p_src = p_src_data;
		uint8_t *p_dest = r_t_dest.p_Data() + n_src_size - 1,
			*p_end = r_t_dest.p_Data() - 1;
		for(uint32_t i = n_primary_index; p_dest != p_end; -- p_dest,
		   i = indices_list[i] + p_buckets[p_src[i]])
			*p_dest = p_src[i];
	}
	// fill output buffer (backwards)

	return true;
}

typedef std::vector<uint32_t>::iterator TULongIter;

class CBurrowsWheelerTransform::CSorter : public CRunable {
protected:
	TULongIter m_p_begin, m_p_end;
	const TBuffer *m_p_src;

public:
	CSorter()
	{}

	CSorter(TULongIter p_begin, TULongIter p_end, const TBuffer &r_t_src)
		:m_p_begin(p_begin), m_p_end(p_end), m_p_src(&r_t_src)
	{}

	virtual void Run()
	{
		std::sort(m_p_begin, m_p_end, CWrapMemCmp(*m_p_src));
		// sort the interval
	}
};

class CBurrowsWheelerTransform::CMerger : public CRunable {
protected:
	TULongIter m_p_out, m_p_begin, m_p_middle, m_p_end;
	std::vector<uint32_t> m_vec;
	const TBuffer *m_p_src;

public:
	CMerger()
	{}

	CMerger(TULongIter p_out, TULongIter p_begin,
		TULongIter p_middle, TULongIter p_end, const TBuffer &r_t_src)
		:m_p_out(p_out), m_p_begin(p_begin), m_p_middle(p_middle), m_p_end(p_end), m_p_src(&r_t_src)
	{}

	virtual void Run()
	{
		std::merge(m_p_begin, m_p_middle, m_p_middle, m_p_end, m_p_out, CWrapMemCmp(*m_p_src));
		// merge somewhere else. my implementation of inplace_merge is not thread safe

		//std::inplace_merge(m_p_begin, m_p_middle, m_p_end, CWrapMemCmp(*m_p_src));
		// sort the interval
	}
};

/*
 *	static bool CBurrowsWheelerTransform::ThreadedEncode(const TBuffer &r_t_in_buffer,
 *		TBuffer &r_t_out_buffer, int n_thread_num)
 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- works in parallel, n_thread_num must be power of two
 *		- returns true on success, false on failure
 *		- note this doesn't work with empty input buffer
 */
bool CBurrowsWheelerTransform::ThreadedEncode(const TBuffer &r_t_src,
	TBuffer &r_t_dest, int n_thread_num)
{
	_ASSERTE(&r_t_src != &r_t_dest);
	_ASSERTE(r_t_src.p_Data());
	// make some assumptions about source buffer

	if(r_t_src.n_Size() > UINT32_MAX - sizeof(uint32_t))
		return false; // overflow
	if(!r_t_dest.Resize(r_t_src.n_Size() + sizeof(uint32_t), false))
		return false;
	// allocate output buffer

	std::vector<uint32_t> indices_list(r_t_src.n_Size());
	std::vector<uint32_t> sorted_list(r_t_src.n_Size());
	if(indices_list.size() < r_t_src.n_Size() ||
	   sorted_list.size() < r_t_src.n_Size())
		return false;
	// allocate list for indices

	std::generate(indices_list.begin(), indices_list.end(), CIota(0));
	// generate sequence of indices

	const int n_max_thread_num = 64;
	// max number of worker threads

	if(n_thread_num > n_max_thread_num)
		return false;

	{
		const uint32_t n_size = r_t_src.n_Size();
		const uint32_t n_part_size = n_size / n_thread_num;
		const uint32_t n_rest_size = n_size % n_thread_num;
		// size of one part for sorting thread and of the rest

		CSorter p_sorter[n_max_thread_num];
		CThread p_thread[n_max_thread_num];

		{
			uint32_t n_begin = 0;
			uint32_t n_cur_size = n_part_size + n_rest_size;
			for(int i = 0; i < n_thread_num; ++ i,
			   n_begin += n_cur_size, n_cur_size = n_part_size) {
				p_sorter[i] = CSorter(indices_list.begin() + n_begin,
					indices_list.begin() + n_begin + n_cur_size, r_t_src);
				p_thread[i].AttachRunable(p_sorter[i]);
				if(!p_thread[i].Start())
					return false;
				// start sorting in threads
			}

			for(int i = 0; i < n_thread_num; ++ i)
				p_thread[i].Stop();
			// wait for threads to finish
		}
		// partially sort list, in n_thread_num threads

		CMerger p_merger[n_max_thread_num / 2];

		for(int n = n_thread_num / 2, m = 1; n != 0; n /= 2, m *= 2) {
			_ASSERTE(!(n % 2) || n == 1); // must be power of two

			uint32_t n_begin = 0;
			uint32_t n_cur_half = n_part_size * m + n_rest_size;
			uint32_t n_cur_size = n_cur_half + n_part_size * m;
			for(int i = 0; i < n; ++ i, n_begin += n_cur_size,
			   n_cur_half = n_part_size * m, n_cur_size = n_part_size * 2 * m) {
				p_merger[i] = CMerger(sorted_list.begin() + n_begin,
					indices_list.begin() + n_begin,
					indices_list.begin() + n_begin + n_cur_half,
					indices_list.begin() + n_begin + n_cur_size, r_t_src);
				p_thread[i].AttachRunable(p_merger[i]);
				if(!p_thread[i].Start())
					return false;
				// start merging in threads
			}
			// merge sub-lists, all merges in one iteration can be executed in parallel

			for(int i = 0; i < n; ++ i)
				p_thread[i].Stop();
			// wait for threads to finish

			indices_list.swap(sorted_list);
		}
		// merge to create sorted list, need n_thread_num / 2 threads
	}
	// sort indices (bottleneck)

	const uint8_t *p_src = r_t_src.p_Data();
	uint8_t *p_dest = r_t_dest.p_Data() + sizeof(uint32_t);
	for(uint32_t i = 0, n = r_t_src.n_Size(); i < n; ++ i, ++ p_dest)
		*p_dest = p_src[(indices_list[i] + n - 1) % n];
	// fill output buffer

	uint32_t n_primary_index = std::find(indices_list.begin(),
		indices_list.end(), 0U) - indices_list.begin();
	_ASSERTE(n_primary_index < r_t_src.n_Size());
	// find primary index

	*(uint32_t*)r_t_dest.p_Data() = n_primary_index;
	// write it to front of the buffer

	return true;
}

/*
 *	static bool CBurrowsWheelerTransform::Encode(const TBuffer &r_t_in_buffer,
 *		TBuffer &r_t_out_buffer)
 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- returns true on success, false on failure
 */
bool CBurrowsWheelerTransform::Encode(const TBuffer &r_t_src, TBuffer &r_t_dest)
{
	_ASSERTE(&r_t_src != &r_t_dest);
	_ASSERTE(r_t_src.p_Data());
	// make some assumptions about source buffer

	if(r_t_src.n_Size() > UINT32_MAX - sizeof(uint32_t))
		return false; // overflow
	if(!r_t_dest.Resize(r_t_src.n_Size() + sizeof(uint32_t), false))
		return false;
	// allocate output buffer

	std::vector<uint32_t> indices_list(r_t_src.n_Size());
	if(indices_list.size() < r_t_src.n_Size())
		return false;
	// allocate list for indices

	std::generate(indices_list.begin(), indices_list.end(), CIota(0));
	// generate sequence of indices

	std::sort(indices_list.begin(), indices_list.end(), CWrapMemCmp(r_t_src));
	// sort indices

	const uint8_t *p_src = r_t_src.p_Data();
	uint8_t *p_dest = r_t_dest.p_Data() + sizeof(uint32_t);
	for(uint32_t i = 0, n = r_t_src.n_Size(); i < n; ++ i, ++ p_dest)
		*p_dest = p_src[(indices_list[i] + n - 1) % n];
	// fill output buffer

	uint32_t n_primary_index = std::find(indices_list.begin(),
		indices_list.end(), 0U) - indices_list.begin();
	_ASSERTE(n_primary_index < r_t_src.n_Size());
	// find primary index

	*(uint32_t*)r_t_dest.p_Data() = n_primary_index;
	// write it to front of the buffer

	return true;
}

/*
 *								=== ~CBurrowsWheelerTransform ===
 */

/*
 *								=== CMoveToFrontTransform ===
 */

/*
 *	static void CMoveToFrontTransform::Decode(const TBuffer &r_t_buffer, int n_algorithm)
 *		- decodes data in r_t_buffer (operates in-sit)
 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
 */
void CMoveToFrontTransform::Decode(TBuffer &r_t_buffer, int n_algorithm)
{
	_Decode(r_t_buffer, r_t_buffer, n_algorithm);
	// decode can work in-sit
}

/*
 *	static void CMoveToFrontTransform::Encode(const TBuffer &r_t_buffer, int n_algorithm)
 *		- encodes data in r_t_buffer (operates in-sit)
 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
 */
void CMoveToFrontTransform::Encode(TBuffer &r_t_buffer, int n_algorithm)
{
	_Encode(r_t_buffer, r_t_buffer, n_algorithm);
	// encode can work in-sit
}

/*
 *	static bool CMoveToFrontTransform::Decode(const TBuffer &r_t_in_buffer,
 *		TBuffer &r_t_out_buffer, int n_algorithm)
 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
 *		- returns true on success, false on failure
 */
bool CMoveToFrontTransform::Decode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm)
{
	if(!r_t_dest.Resize(r_t_src.n_Size(), false))
		return false;
	// allocate output buffer

	_Decode(r_t_src, r_t_dest, n_algorithm);
	// decode

	return true;
}

/*
 *	static bool CMoveToFrontTransform::Encode(const TBuffer &r_t_in_buffer,
 *		TBuffer &r_t_out_buffer, int n_algorithm)
 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
 *		- returns true on success, false on failure
 */
bool CMoveToFrontTransform::Encode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm)
{
	if(!r_t_dest.Resize(r_t_src.n_Size(), false))
		return false;
	// allocate output buffer

	_Encode(r_t_src, r_t_dest, n_algorithm);
	// encode

	return true;
}

/*
 *	static bool CMoveToFrontTransform::_Decode(const TBuffer &r_t_src,
 *		TBuffer &r_t_dest, int n_algorithm)
 *		- decodes data from r_t_src, outputs to r_t_dest which
 *		  must be allocated to the same size as r_t_src
 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
 */
void CMoveToFrontTransform::_Decode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm)
{
	_ASSERTE(r_t_src.n_Size() == r_t_dest.n_Size());
	_ASSERTE(r_t_src.p_Data() && r_t_dest.p_Data());
	// make some assumptions about input / output buffers

	const int n_max_word_value = 1 << (8 * sizeof(uint8_t));
	// number of combinations in a single word

	uint8_t p_data[n_max_word_value];
	for(int i = 0; i < n_max_word_value; ++ i)
		p_data[i] = (uint8_t)i;
	// start with buffer of successive numbers

	if(n_algorithm == algo_MTF) {
		const uint8_t *p_src = r_t_src.p_Data(),
			*p_end = r_t_src.p_Data() + r_t_src.n_Size();
		for(uint8_t *p_dest = r_t_dest.p_Data(); p_src != p_end; ++ p_src, ++ p_dest) {
			uint8_t n_index = *p_src;
			if(!n_index)
				*p_dest = p_data[0];
			else {
				int n_data = p_data[n_index];
				// get index in the array

				for(int i = n_index; i > 0; -- i)
					p_data[i] = p_data[i - 1];
				// shift the array (simulate encoder)

				p_data[0] = n_data;
				// move our byte to the front

				*p_dest = n_data;
				// output index where the byte has been found
			}
		}
	} else /*if(n_algorithm == algo_MTF_1)*/ {
		const uint8_t *p_src = r_t_src.p_Data(),
			*p_end = r_t_src.p_Data() + r_t_src.n_Size();
		for(uint8_t *p_dest = r_t_dest.p_Data(); p_src != p_end; ++ p_src, ++ p_dest) {
			uint8_t n_index = *p_src;
			if(!n_index)
				*p_dest = p_data[0];
			else {
				int n_data = p_data[n_index];
				// get index in the array

				if(n_index == 1) {
					for(int i = n_index; i > 0; -- i)
						p_data[i] = p_data[i - 1];
					// shift the array (simulate encoder)

					p_data[0] = n_data;
					// move our byte to the front
				} else {
					for(int i = n_index; i > 1; -- i)
						p_data[i] = p_data[i - 1];
					// shift the array (simulate encoder)

					p_data[1] = n_data;
					// move our byte to second place
				}

				*p_dest = n_data;
				// output index where the byte has been found
			}
		}
	}
	// decode
}

/*
 *	static bool CMoveToFrontTransform::_Encode(const TBuffer &r_t_src,
 *		TBuffer &r_t_dest, int n_algorithm)
 *		- encodes data from r_t_src, outputs to r_t_dest which
 *		  must be allocated to the same size as r_t_src
 *		- n_algorithm is MTF algorithm, one of algo_MTF, algo_MTF_1
 */
void CMoveToFrontTransform::_Encode(const TBuffer &r_t_src, TBuffer &r_t_dest, int n_algorithm)
{
	_ASSERTE(r_t_src.n_Size() == r_t_dest.n_Size());
	_ASSERTE(r_t_src.p_Data() && r_t_dest.p_Data());
	// make some assumptions about input / output buffers

	const int n_max_word_value = 1 << (8 * sizeof(uint8_t));
	// number of combinations in a single word

	uint8_t p_indices[n_max_word_value];
	for(int i = 0; i < n_max_word_value; ++ i)
		p_indices[i] = (uint8_t)i;
	// start with buffer of successive numbers

	if(n_algorithm == algo_MTF) {
		const uint8_t *p_src = r_t_src.p_Data(),
			*p_end = r_t_src.p_Data() + r_t_src.n_Size();
		for(uint8_t *p_dest = r_t_dest.p_Data(); p_src != p_end; ++ p_src, ++ p_dest) {
			uint8_t n_data = *p_src;
			if(n_data == p_indices[0])
				*p_dest = 0; // we wish this would happen a lot
			else {
				int n_index = 0;
				for(;; ++ n_index) {
					_ASSERTE(n_index < n_max_word_value);
					if(p_indices[n_index] == n_data)
						break;
				}
				// find it in array

				for(int i = n_index; i > 0; -- i)
					p_indices[i] = p_indices[i - 1];
				// shift the array

				p_indices[0] = n_data;
				// move our byte to the front

				*p_dest = n_index;
				// output index where the byte has been found
			}
		}
	} else /*if(n_algorithm == algo_MTF_1)*/ {
		_ASSERTE(n_algorithm == algo_MTF_1);
		const uint8_t *p_src = r_t_src.p_Data(),
			*p_end = r_t_src.p_Data() + r_t_src.n_Size();
		for(uint8_t *p_dest = r_t_dest.p_Data(); p_src != p_end; ++ p_src, ++ p_dest) {
			uint8_t n_data = *p_src;
			if(n_data == p_indices[0])
				*p_dest = 0; // we wish this would happen a lot
			else {
				int n_index = 0;
				for(;; ++ n_index) {
					_ASSERTE(n_index < n_max_word_value);
					if(p_indices[n_index] == n_data)
						break;
				}
				// find it in array

				if(n_index == 1) {
					for(int i = n_index; i > 0; -- i)
						p_indices[i] = p_indices[i - 1];
					// shift the array

					p_indices[0] = n_data;
					// move our byte to the front
				} else {
					for(int i = n_index; i > 1; -- i)
						p_indices[i] = p_indices[i - 1];
					// shift the array

					p_indices[1] = n_data;
					// move our byte to second place
				}

				*p_dest = n_index;
				// output index where the byte has been found
			}
		}
	}
	// encode
}

/*
 *								=== ~CMoveToFrontTransform ===
 */

/*
 *								=== CRunLengthCodec ===
 */

/*
 *	static bool CRunLengthCodec::Decode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- returns true on success, false on failure
 */
bool CRunLengthCodec::Decode(const TBuffer &r_t_src, TBuffer &r_t_dest)
{
	uint8_t *p_output = r_t_dest.p_Data(),
		*p_end = r_t_dest.p_Data() + r_t_dest.n_Size();
	// output buffer (size will change dynamically)

	for(const uint8_t *p_src = r_t_src.p_Data(),
	   *p_end2 = r_t_src.p_Data() + r_t_src.n_Size(); p_src != p_end2;) {
		uint8_t n_code = *p_src;
		bool b_compressed = n_code & 1;
		int n_run_length = (n_code >> 1) + 1;
		// determine compression and run length

		if(p_src + ((b_compressed)? 2 : 1 + n_run_length) > p_end2)
			return false;
		// buffer overrun, invalid input data

		++ p_src;
		// skip code

		if(p_output + n_run_length > p_end) {
			uint32_t n_off = p_output - r_t_dest.p_Data();
			if(!r_t_dest.Grow(n_run_length))
				return false;
			p_output = r_t_dest.p_Data() + n_off;
			p_end = r_t_dest.p_Data() + r_t_dest.n_Size();
		}
		// make sure there's enough space

		if(b_compressed) {
			uint8_t n_data = *p_src ++;
			for(const uint8_t *p_end2 = p_output + n_run_length;
			   p_output != p_end2; ++ p_output)
				*p_output = n_data;
			// replicate the same byte n_run_length times
		} else {
			for(const uint8_t *p_end2 = p_output + n_run_length;
			   p_output != p_end2; ++ p_output, ++ p_src)
				*p_output = *p_src;
			// copy n_run_length bytes
		}
		// read data

		_ASSERTE(p_output <= p_end);
		_ASSERTE(p_end == r_t_dest.p_Data() + r_t_dest.n_Size());
		// make sure we don't cross buffer boundaries
	}
	// IRLE loop

	if(!r_t_dest.Resize(p_output - r_t_dest.p_Data())) {
		_ASSERTE(0);
		return false;
	}
	// shrink the buffer to it's final length

	return true;
}

/*
 *	static bool CRunLengthCodec::Encode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- returns true on success, false on failure
 */
bool CRunLengthCodec::Encode(const TBuffer &r_t_src, TBuffer &r_t_dest)
{
	const bool b_zero_rle = false;
	const int n_min_run_length = 2;
	const int n_max_repeats = (1 << (8 * sizeof(uint8_t) - 1)) + 1;
	// compressor config

	uint8_t *p_output = r_t_dest.p_Data(),
		*p_end = r_t_dest.p_Data() + r_t_dest.n_Size();
	// output buffer (size will change dynamically)

	for(const uint8_t *p_src = r_t_src.p_Data(),
	   *p_end2 = r_t_src.p_Data() + r_t_src.n_Size(); p_src != p_end2;) {
		int n_remains = min(p_end2 - p_src, n_max_repeats);
		// see how much bytes do we have left

		bool b_compress = n_remains > 1 &&
			(!b_zero_rle || *p_src == 0) && *p_src == *(p_src + 1);
		// are we gonna try to compress?

		int n_run_length;
		do {
			if(b_compress) {
				n_run_length = 1;
				// minimal run length without entering the loop

				for(int i = 2; i < n_remains; ++ i) {
					n_run_length = i;
					if(*p_src != *(p_src + i))
						break;
				}
				// determine run length

				if(n_run_length < n_min_run_length) {
					b_compress = false;
					// won't compress too short sequences
				} else {
					break;
					// skip the other branch
				}
			}
			// determine compressed run length

			if(!b_compress) { // not else if!
				n_run_length = 1;
				// minimal run length without entering the loop

				for(int i = 2, n_possible_run = 0; i < n_remains; ++ i) {
					if(*(p_src + i) == *(p_src + i - 1)) {
						if(++ n_possible_run >= n_min_run_length) {
							n_run_length = i - n_possible_run;
							break;
						}
					} else {
						n_possible_run = 0;
						n_run_length = i;
					}
				}
				// find first repeating sequence of minimal length
			}
			// determine uncompressed run length
		} while(0);
		// determine run length

		int n_add = (b_compress)? 2 : 1 + n_run_length;
		if(p_output + n_add > p_end) {
			uint32_t n_off = p_output - r_t_dest.p_Data();
			if(!r_t_dest.Grow(n_add))
				return false;
			p_output = r_t_dest.p_Data() + n_off;
			p_end = r_t_dest.p_Data() + r_t_dest.n_Size();
		}
		// make sure there's enough space

		_ASSERTE(n_run_length > 0);
		*p_output ++ = ((b_compress)? 1 : 0) | ((n_run_length - 1) << 1);
		// write run length

		if(b_compress) {
			*p_output ++ = *p_src;
			p_src += n_run_length;
		} else {
			for(int i = 0; i < n_run_length; ++ i, ++ p_output, ++ p_src)
				*p_output = *p_src;
		}
		// write data

		_ASSERTE(p_output <= p_end);
		_ASSERTE(p_end == r_t_dest.p_Data() + r_t_dest.n_Size());
		// make sure we don't cross buffer boundaries
	}
	// RLE loop

	if(!r_t_dest.Resize(p_output - r_t_dest.p_Data())) {
		_ASSERTE(0);
		return false;
	}
	// shrink the buffer to it's final length

	return true;
}

/*
 *								=== ~CRunLengthCodec ===
 */

/*
 *								=== CHuffmanCodec ===
 */

/*
 *	CHuffmanCodec::TFrequency
 *		- structure, holding symbol, it's frequency and eventualy also it's code
 */
struct CHuffmanCodec::TFrequency {
	uint8_t n_symbol; // valid for leaf nodes only
	uint32_t n_frequency;

	int n_code_word;
	int n_code_length;

	TFrequency(uint8_t n_sym)
		:n_symbol(n_sym), n_frequency(0), n_code_word(0), n_code_length((unsigned int)-1)
	{}

	TFrequency(uint8_t n_sym, uint32_t n_freq)
		:n_symbol(n_sym), n_frequency(n_freq), n_code_word(0), n_code_length((unsigned int)-1)
	{}

	inline bool operator <(const TFrequency &r_t_freq) const // lower frequency comparison (reversed)
	{
		return n_frequency > r_t_freq.n_frequency;
	}

	inline bool operator ==(uint8_t n_sym) const // symbol equality comparison
	{
		return n_symbol == n_sym;
	}

	inline bool operator <(uint8_t n_sym) const // symbol equality comparison
	{
		return n_symbol < n_sym;
	}
};

/*
 *	inline bool CHuffmanCodec::FindZeroFreq(const TFrequency &r_t_freq)
 *		- zero frequency predicate
 */
inline bool CHuffmanCodec::FindZeroFreq(const TFrequency &r_t_freq)
{
	return !r_t_freq.n_frequency;
}

/*
 *	inline bool CHuffmanCodec::CompareSymbol(const TFrequency &r_freq_a,
 *		const TFrequency &r_freq_b)
 *		- less-than ordering for frequency symbol comparison
 */
inline bool CHuffmanCodec::CompareSymbol(const TFrequency &r_freq_a, const TFrequency &r_freq_b)
{
	return r_freq_a.n_symbol < r_freq_b.n_symbol;
}

struct CHuffmanCodec::TNode {
	uint32_t n_frequency;
	const TNode *p_left, *p_right;

	TNode(uint32_t n_freq = 0, const TNode *p_l = 0, const TNode *p_r = 0)
		:n_frequency(n_freq), p_left(p_l), p_right(p_r)
	{}

	bool b_Leaf() const
	{
		_ASSERTE(!p_left == !p_right);
		return !p_left && !p_right;
	}

	inline bool operator <(const TNode &r_t_node) const // lower frequency comparison (reversed)
	{
		return n_frequency > r_t_node.n_frequency;
	}
};

bool CHuffmanCodec::CreateHuffmanTree(std::vector<TFrequency> &r_freq_list, int *p_code_num)
{
	if(r_freq_list.size() > (1 << max_CodeBitNum))
		return false;
	// too much symbols

	std::sort(r_freq_list.begin(), r_freq_list.end());
	// sort by symbol frequencies

	r_freq_list.erase(std::find_if(r_freq_list.begin(),
		r_freq_list.end(), FindZeroFreq), r_freq_list.end());
	// erase any symbols with zero frequencies

	std::vector<TNode> original_nodes;
	original_nodes.reserve(r_freq_list.size() * 2 - 1);
	if(original_nodes.capacity() < r_freq_list.size() * 2 - 1)
		return false;
	// create list of all nodes in the tree (including non-leaf nodes)

	for(int i = 0, n = r_freq_list.size(); i < n; ++ i)
		original_nodes.push_back(TNode(r_freq_list[i].n_frequency));
	// add nodes holding original symbol frequencies

	original_nodes.insert(original_nodes.end(), r_freq_list.size() - 1, TNode());
	// add new nodes which will be used as non-leaf nodes

	std::vector<const TNode*> q1, q2;
	// have two queues

	q1.reserve(r_freq_list.size());
	if(q1.capacity() < r_freq_list.size())
		return false;
	q2.reserve((r_freq_list.size() + 1) / 2);
	if(q2.capacity() < (r_freq_list.size() + 1) / 2)
		return false;
	// make sure there's plenty of space in both of them

	bool b_restart = false;
	for(;;) {
		for(int i = 0, n = r_freq_list.size(); i < n; ++ i)
			q1.push_back(&original_nodes[i]);
		// add all original nodes to the first queue

		int n_free_node = r_freq_list.size();
		const int n_end_node = original_nodes.size();
		// index to the first free node and one node past the ond of the list

		while(q1.size() + q2.size() > 1) {
			const TNode *p_node[2];
			for(int i = 0; i < 2; ++ i) {
				if(!q2.size() || (q1.size() && q1.back()->n_frequency < q2.back()->n_frequency)) {
					p_node[i] = q1.back();
					q1.erase(q1.end() - 1, q1.end());
					// node in the first queue has smaller freq or is the only left
				} else {
					p_node[i] = q2.back();
					q2.erase(q2.end() - 1, q2.end());
					// node in second queue has smaller freq or is the only left
				}
			}
			// choose two nodes to merge

			_ASSERTE(n_free_node != n_end_node);
			TNode *p_new = &original_nodes[n_free_node ++];
			// get some unused node

			_ASSERTE(p_node[0]->n_frequency <= UINT32_MAX - p_node[1]->n_frequency);
			*p_new = TNode(p_node[0]->n_frequency + p_node[1]->n_frequency, p_node[0], p_node[1]);
			// assign frequency and children

			q2.insert(q2.begin(), p_new);
			_ASSERTE(q2.size() <= (r_freq_list.size() + 1) / 2);
			// put it to second queue
		}
		// create huffman tree

		_ASSERTE(n_free_node == n_end_node); // used all the nodes in the tree
		_ASSERTE(!q1.size() && q2.size() == 1); // root node is in q2

		memset(p_code_num, 0, max_CodeBitNum * sizeof(int));
		// clear the code counts array

		int n_level = 0;
		while(q1.size() || q2.size()) {
			std::vector<const TNode*> &r_qa = (n_level & 1)? q1 : q2;
			std::vector<const TNode*> &r_qb = (n_level & 1)? q2 : q1;

			for(int i = 0, n = r_qa.size(); i < n; ++ i) {
				const TNode *p_cur = r_qa[i];
				if(p_cur->b_Leaf()) {
					++ p_code_num[max(0, n_level - 1)];
				} else {
					_ASSERTE(p_cur->p_left && p_cur->p_right);
					r_qb.push_back(p_cur->p_left);
					r_qb.push_back(p_cur->p_right);
				}
			}
			r_qa.erase(r_qa.begin(), r_qa.end());
			// alternately move whole tree levels between q1 and q2

			if(++ n_level > max_CodeBitNum) {
				b_restart = true;
				break;
			}
		}
		// determine number of symbols with distinct lengths

		if(!b_restart) {
			for(int i = 0, j = 0, n_code_word = 0, n_prev_code_length = 0,
			   n_symbol_num = r_freq_list.size(); j < max_CodeBitNum && n_symbol_num; ++ j) {
				int n_code_length = j + 1;
				if(i && n_code_length != n_prev_code_length) {
					if(n_code_word > (INT_MAX >> (n_code_length - n_prev_code_length))) {
						b_restart = true;
						break;
						// overflow
					}
					n_code_word <<= n_code_length - n_prev_code_length;
				}
				for(int k = 0; k < p_code_num[j]; ++ k, ++ i) {
					TFrequency &r_freq = r_freq_list[i];
					r_freq.n_code_word = n_code_word ++;
					if(n_code_word < 0) {
						b_restart = true;
						break;
						// overflow
					}
					r_freq.n_code_length = n_code_length;
					if(r_freq.n_code_word >> r_freq.n_code_length) {
						b_restart = true;
						break;
						// code word is longer, we need to reassign some of the weights
					}
				}
				if(b_restart)
					break;
				n_symbol_num -= p_code_num[j]; // we'we processed this much symbols
				if(p_code_num[j])
					n_prev_code_length = n_code_length;
			}
		}
		// assign code words, fill-in code lengths

		if(b_restart) {
			for(int i = 0, n = r_freq_list.size(); i < n; ++ i) {
				TNode &r_node = original_nodes[i];
				if(!(r_node.n_frequency /= 2))
					r_node.n_frequency = 1;
			}
			// round node frequencies

			std::sort(original_nodes.begin(),
				original_nodes.begin() + r_freq_list.size());
			// sort leaf nodes by frequency

			q1.erase(q1.begin(), q1.end());
			q2.erase(q2.begin(), q2.end());
			_ASSERTE(q1.capacity() >= r_freq_list.size());
			_ASSERTE(q2.capacity() >= (r_freq_list.size() + 1) / 2);
			// make sure they're still allocated, and empty

			b_restart = false;
		} else
			break;
		// we need to modify codes so it would get built
	}

	return true;
}

/*
 *	static bool CHuffmanCodec::Decode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
 *		- decodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- returns true on success, false on failure
 */
bool CHuffmanCodec::Decode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
{
	const int n_max_word_value = 1 << (8 * sizeof(uint8_t));
	// number of combinations in a single word

	int p_code_num[max_CodeBitNum];
	// numbers of codes of different lengths

	uint8_t p_symbol[n_max_word_value];
	// symbols in the ascending order of associated codes

	if(r_t_in_buffer.n_Size() < sizeof(p_code_num) + sizeof(uint32_t))
		return false;
	// least possible size of input

	const uint8_t *p_input = r_t_in_buffer.p_Data();
	// input buffer pointer

	uint32_t n_unpack_length = *(uint32_t*)p_input;
	p_input += sizeof(uint32_t);
	// get size of uncompressed data

	memcpy(p_code_num, p_input, sizeof(p_code_num));
	p_input += sizeof(p_code_num);
	// get numbers of codes

	int p_min_code[max_CodeBitNum];
	int p_max_code[max_CodeBitNum];
	int p_table_offset[max_CodeBitNum];
	// minimal / maximal codes for given lengths and offsets to symbol table

	int n_symbol_num = 0;
	for(int i = 0, n_code_word = 0, n_prev_code_length = 0; i < max_CodeBitNum; ++ i) {
		int n_code_length = i + 1; // used as difference value, don't need + 1
		if(n_symbol_num && n_code_length != n_prev_code_length)
			n_code_word <<= n_code_length - n_prev_code_length;
		p_min_code[i] = n_code_word;
		n_code_word += p_code_num[i];
		p_max_code[i] = n_code_word;
		p_table_offset[i] = n_symbol_num;
		n_symbol_num += p_code_num[i];
		if(p_code_num[i])
			n_prev_code_length = n_code_length;
	}
	// calculate number of symbols, table indices and min / max code values

	if(r_t_in_buffer.n_Size() < sizeof(p_code_num) + sizeof(uint32_t) + n_symbol_num)
		return false;
	// check size of input

	memcpy(p_symbol, p_input, n_symbol_num * sizeof(uint8_t));
	p_input += n_symbol_num;
	// copy symbols

	if(!r_t_out_buffer.Resize(n_unpack_length, false))
		return false;
	// alloc output buffer

	uint8_t n_byte = 0;
	int n_bit_num = 0;
	for(uint8_t *p_dest = r_t_out_buffer.p_Data(),
	   *p_end = r_t_out_buffer.p_Data() + r_t_out_buffer.n_Size(); p_dest != p_end; ++ p_dest) {
		int n_code = 0;

		for(int i = 0; /*i < max_CodeBitNum*/; /*++ i*/) {
			if(!n_bit_num) {
				if(p_input == r_t_in_buffer.p_Data() + r_t_in_buffer.n_Size())
					return false;
				// not enough data to decompress

				n_byte = *p_input ++;
				n_bit_num = 7;
			} else
				-- n_bit_num;
			int n_bit = n_byte >> 7;
			n_byte <<= 1;
			// get a single bit from input stream

			n_code <<= 1;
			n_code |= n_bit;
			// add bit to code

			if(n_code >= p_min_code[i] && n_code < p_max_code[i]) {
				*p_dest = p_symbol[n_code - p_min_code[i] + p_table_offset[i]];
				break;
			}
			// see if it's valid code for this bit length

			if(++ i == max_CodeBitNum)
				return false; // invalid code
		}
	}
	// decode data

	_ASSERTE(p_input == r_t_in_buffer.p_Data() + r_t_in_buffer.n_Size());
	// make sure we've read the whole input buffer

	return true;
}

/*
 *	static bool CHuffmanCodec::Encode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
 *		- encodes data from r_t_in_buffer, outputs to r_t_out_buffer (can be empty)
 *		- returns true on success, false on failure
 */
bool CHuffmanCodec::Encode(const TBuffer &r_t_in_buffer, TBuffer &r_t_out_buffer)
{
	const unsigned int n_max_word_value = 1 << (8 * sizeof(uint8_t));
	// number of combinations in a single word

	std::vector<TFrequency> freq_list;
	freq_list.reserve(n_max_word_value);
	if(freq_list.capacity() < n_max_word_value)
		return false;
	// buffer with symbol frequencies

	for(unsigned int i = 0; i < n_max_word_value; ++ i)
		freq_list.push_back(TFrequency(i));
	// set symbols

	{
		const uint8_t *p_src = r_t_in_buffer.p_Data(),
			*p_end = r_t_in_buffer.p_Data() + r_t_in_buffer.n_Size();
		for(; p_src != p_end; ++ p_src) {
			++ freq_list[*p_src].n_frequency;
			// calculate frequency

			_ASSERTE(freq_list[*p_src].n_frequency);
			// owerflow shouldn't occur since sum of all frequencies
			// is buffer length which is uint32_t as well
		}
	}
	// calculate symbol frequencies

	int p_code_num[max_CodeBitNum] = {0};
	if(!CreateHuffmanTree(freq_list, p_code_num))
		return false;
	// create huffman tree

	int n_header_size = sizeof(uint32_t) +
		sizeof(p_code_num) + freq_list.size();
	// we're going to need to store the huffman tree
	// (numbers of codes with distinct lengths and symbols)

	if(!r_t_out_buffer.Resize(n_header_size, false))
		return false;
	// make sure there's enough space for header

	uint8_t *p_output = r_t_out_buffer.p_Data(),
		*p_end = r_t_out_buffer.p_Data() + n_header_size;
	// output buffer (size will change dynamically)

	*(uint32_t*)p_output = r_t_in_buffer.n_Size();
	p_output += sizeof(uint32_t);
	// write size of decompressed stream so decompressor can work in a single pass

	memcpy(p_output, p_code_num, sizeof(p_code_num));
	p_output += sizeof(p_code_num);
	for(int i = 0, n = freq_list.size(); i < n; ++ i, ++ p_output)
		*p_output = freq_list[i].n_symbol;
	// write numbers of codes of different lengths and associated symbols

	std::sort(freq_list.begin(), freq_list.end(), CompareSymbol);
	// sort by symbols so we can search using lower_bound (binary search)

	{
		uint8_t m_n_byte = 0;
		int m_n_bit_num = 0;
		// bit writer

		const uint8_t *p_src = r_t_in_buffer.p_Data(),
			*p_end2 = r_t_in_buffer.p_Data() + r_t_in_buffer.n_Size();
		for(; p_src != p_end2; ++ p_src) {
			const TFrequency &r_freq = *std::lower_bound(freq_list.begin(),
				freq_list.end(), *p_src);
			_ASSERTE(r_freq.n_symbol == *p_src);
			// find symbol in huffman tree

			int n_bit_num = r_freq.n_code_length;
			int n_value = r_freq.n_code_word;
			for(-- n_bit_num; n_bit_num >= 0; -- n_bit_num) {
				int n_bit = (n_value >> n_bit_num) & 1;

				m_n_byte <<= 1;
				m_n_byte |= n_bit;
				if(m_n_bit_num == 7) {
					m_n_bit_num = 0;
					if(p_output == p_end) {
						uint32_t n_off = p_output - r_t_out_buffer.p_Data();
						if(!r_t_out_buffer.Grow(1))
							return false;
						p_output = r_t_out_buffer.p_Data() + n_off;
						p_end = r_t_out_buffer.p_Data() + r_t_out_buffer.n_Size();
					}
					*p_output ++ = m_n_byte;
				} else
					m_n_bit_num ++;
			}
			// write symbol as series of bits
		}

		if(m_n_bit_num) {
			m_n_byte <<= 8 - m_n_bit_num;
			// padd with 0-s

			if(p_output == p_end) {
				uint32_t n_off = p_output - r_t_out_buffer.p_Data();
				if(!r_t_out_buffer.Grow(1))
					return false;
				p_output = r_t_out_buffer.p_Data() + n_off;
				p_end = r_t_out_buffer.p_Data() + r_t_out_buffer.n_Size();
			}
			*p_output ++ = m_n_byte;
		}
		// output any leftover bits
	}
	// compress

	if(!r_t_out_buffer.Resize(p_output - r_t_out_buffer.p_Data())) {
		_ASSERTE(0);
		return false;
	}
	// shrink the buffer to it's final length

	return true;
}

/*
 *								=== ~CHuffmanCodec ===
 */
