/*
								+---------------------------------+
								|                                 |
								|   ***   3ds file loader  ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|            Load3ds.h            |
								|                                 |
								+---------------------------------+
*/

/*
 *	2006-06-05
 *
 *	rewritten based on the old "C" 3ds loader
 *
 *	2007-06-04
 *
 *	fixed !fread typo in ReadTrackColor()
 *
 */

#ifndef LOAD_3DS_H

/*
 *	T3DS_Vertex
 *		struktura s informacemi o geometrickych vrcholech sveta
 */
struct T3DS_Vertex {
	int n_flag;

    Vector3f v_pos;
	Vector3f v_texcoord;

    Vector3f v_normal;
};

/*
 *	TMaterial
 *		struktura s informacemi o pouzitych materialech
 */
struct TMaterial {
    int n_id;
    char *p_s_name;
	int n_ogl_id;

	Vector3f v_ambient;
	Vector3f v_diffuse;
	Vector3f v_specular;
	float f_self_illum;
};

/*
 *	TLight
 *		struktura s informacemi o svetlech (pri radiosite takto svetla nejsou definovany)
 */
struct TLight {
    int n_id;
    char *p_s_name;

    int b_spot;
    int b_off;
    int b_cast_shadows;

    Vector3f v_position;
    Vector3f v_target;

    float f_hotspot;
    float f_faloff;

    float f_inner_range;
    float f_outer_range;
    float f_multiplier;

    Vector3f v_color;
};

/*
 *	TFace
 *		struktura, uchovavajici topologii trojuhelnika a jeste par dalsich udaju
 */
struct TFace {
    int n_id, n_used;
    int n_material;

	int n_group;
	char b_original;
	char b_neg_plane;

    T3DS_Vertex *p_vertex[3];

	Vector3f v_center;
	float f_radius;

    Plane3f t_normal;
};

/*
 *	TPosFrame
 *		animacni ramec s pozici (bod)
 */
struct TPosFrame {
	float f_time;
	Vector3f v_position;
	float f_tension, f_continuity, f_bias;
};

/*
 *	TRotFrame
 *		animacni ramec s rotaci (quaternion)
 */
struct TRotFrame {
	float f_time;
	Vector4f v_rotation; // quaternion(x y z w)
	float f_tension, f_continuity, f_bias;
};

/*
 *	TObject
 *		sdruzuje vrcholy a trojuhelniky objektu spolu s animacnimi strukturami
 */
struct TObject {
    int n_id;
    char *p_s_name;

    int n_vertex_num;
    int n_face_num;

    TFace *p_face;
    T3DS_Vertex *p_vertex;

    Matrix4f m_object;

    int n_parent;
    int n_pos_num;
    int n_rot_num;

	TPosFrame *p_pos;
	TRotFrame *p_rot;

	void Calc_FaceNormals();
	void Calc_VertexNormals(); // need to Calc_FaceNormals() first

	bool WeldVertices();
};

/*
 *	TCamera
 *		obsahuje informace o kamerach (nakreslenych ve 3D studiu)
 */
struct TCamera {
    int n_id;
    char *p_s_name;

    Matrix4f m_camera;

	Vector3f v_position;
	Vector3f v_target;
    float f_bank_angle;
    float f_focus; 

    int n_target_num;
    int n_position_num;
	
	TPosFrame *p_position;
	TPosFrame *p_target;

	void CalcMatrix();
};

/*
 *	TWorld
 *		struktura linearniho R3 sveta
 */
struct TWorld {
    int n_start_frame;
    int n_end_frame;

    int n_object_num;
    int n_camera_num;
    int n_light_num;
    int n_material_num;

    TObject *p_object;
    TCamera *p_camera;
    TLight *p_light;
    TMaterial *p_material;

	TWorld();

	void Free();
};

/*
 *	C3DSLoader
 *		trida, umoznujici nahravani TWorld ze souboru
 *		(sama je TWorld, ale nikdy neobsahuje zadna data)
 */
class C3DSLoader : protected TWorld {
protected:
	Vector3f m_v_color;
	float m_f_percentage;

	int m_n_node_object_index;
	int m_n_face_num;

	char m_p_s_name[64];
	char m_p_s_track_name[64];

	int m_n_track_type, m_n_camera_index, m_n_object_index;
	int m_n_material_index, m_n_element_index;
	int m_n_light_index;

	int m_b_count_entites;

	struct TChunkHeader { 
		unsigned short n_id;
		unsigned long n_len;
	};

	struct TChunkName {
		unsigned short n_id;
		bool b_subchunk;
		bool (*function)(C3DSLoader *p_this, FILE *fp, long n_pos);
	};

	static const TChunkName p_chunk_name_list[];

	enum {
		chunk_Version           = 0x0002,
		chunk_Color_F           = 0x0010,
		chunk_Color_24          = 0x0011,
		chunk_Lin_Color24       = 0x0012,
		chunk_Lin_Color_F       = 0x0013,
		chunk_N_Percentage      = 0x0030,
		chunk_F_Percentage      = 0x0031,
		chunk_Master_Scale      = 0x0100,
		chunk_BitMap            = 0x1100,
		chunk_Use_BitMap        = 0x1101,
		chunk_Bkgnd_Color       = 0x1200,
		chunk_Ambient_Color     = 0x2100,
		chunk_Layer_Fog         = 0x2302,
		chunk_MeshData          = 0x3D3D,
		chunk_MLib_Magic        = 0x3DAA,
		chunk_Mesh_Version      = 0x3D3E,
		chunk_Named_Object      = 0x4000,
		chunk_Triangle_Object        = 0x4100,
		chunk_Point_List        = 0x4110,
		chunk_Point_Flags       = 0x4111,
		chunk_Face_List         = 0x4120,
		chunk_Material_Group    = 0x4130,
		chunk_Tex_VertexS         = 0x4140,
		chunk_Smooth_Group      = 0x4150,
		chunk_Mesh_Matrix       = 0x4160,
		chunk_Mesh_Color        = 0x4165,
		chunk_Texture_Info      = 0x4170,
		chunk_Mesh_BoxMap       = 0x4190,
		chunk_Light             = 0x4600,
		chunk_SpotLight         = 0x4610,
		chunk_Light_Off         = 0x4620,
		chunk_Light_Shadowed    = 0x4627,
		chunk_Light_Attent      = 0x4625,
		chunk_Light_Range       = 0x4655,
		chunk_Light_Ray_Bias    = 0x4658,
		chunk_Light_In_Range    = 0x4659,
		chunk_Light_Out_Range   = 0x465A,
		chunk_Light_Multiplier  = 0x465B,
		chunk_Camera            = 0x4700,
		chunk_Main_Magic        = 0x4D4D,
		chunk_Hierarchy         = 0x4F00,
		chunk_Viewport_Layout   = 0x7001,
		chunk_Viewport_Size     = 0x7020,
		chunk_Material          = 0xAFFF,
		chunk_Material_Name     = 0xA000,
		chunk_Material_Ambient  = 0xA010,
		chunk_Material_Diffuse  = 0xA020,
		chunk_Material_Specular = 0xA030,
		chunk_Material_Shininess    = 0xA040,
		chunk_Material_ShinStrength = 0xA041,
		chunk_Material_Transparency = 0xA050,
		chunk_Material_XpFall   = 0xA052,
		chunk_Material_RefBlur  = 0xA053,
		chunk_Material_SelfIllum    = 0xA084,
		chunk_Material_DoubleSided = 0xA081,
		chunk_Material_TransAdd = 0xA083,
		chunk_Material_Wire_On   = 0xA085,
		chunk_Material_WireSize = 0xA087,
		chunk_Material_Soften   = 0xA08C,
		chunk_Material_Shading  = 0xA100,
		chunk_Material_Texture  = 0xA200,
		chunk_Material_Reflect  = 0xA220,
		chunk_Material_BumpMap  = 0xA230,
		chunk_Material_BumpAmount   = 0xA252,
		chunk_Material_MapName  = 0xA300,
		chunk_Material_MapFlags = 0xA351,
		chunk_Material_MapBlur  = 0xA353,
		chunk_Material_MapUScale    = 0xA354,
		chunk_Material_MapVScale    = 0xA356,
		chunk_Material_MapUOffset   = 0xA358,
		chunk_Material_MapVOffset   = 0xA35A,
		chunk_Material_MapRotAngle  = 0xA35C,
		chunk_KF_Data           = 0xB000,
		chunk_Ambient_Key       = 0xB001,
		chunk_Track_Info        = 0xB002,
		chunk_Track_Camera      = 0xB003,
		chunk_Track_Cam_Target  = 0xB004,
		chunk_Track_Light       = 0xB005,
		chunk_Track_Light_Target    = 0xB006,
		chunk_Track_SpotLight   = 0xB007,
		chunk_KF_Seg            = 0xB008,
		chunk_KF_CurTime        = 0xB009,
		chunk_KF_HDR            = 0xB00A,
		chunk_Track_Node_Name   = 0xB010,
		chunk_Instance_Name     = 0xB011,
		chunk_PreScale          = 0xB012,
		chunk_Track_Pivot       = 0xB013,
		chunk_BoundBox          = 0xB014,
		chunk_Morph_Smooth      = 0xB015,
		chunk_Track_Pos         = 0xB020,
		chunk_Track_Rotate      = 0xB021,
		chunk_Track_Scale       = 0xB022,
		chunk_Track_FOV         = 0xB023,
		chunk_Track_Roll        = 0xB024,
		chunk_Track_Color       = 0xB025,
		chunk_Track_Morph       = 0xB026,
		chunk_Track_Hotspot     = 0xB027,
		chunk_Track_Falloff     = 0xB028,
		chunk_Track_Hide        = 0xB029,
		chunk_Node_Id           = 0xB030,
		chunk_C_Magic           = 0xC23D
	};

	enum {
		Unknown = 0,
		Camera_Misc,
		Camera_Tar,
		Camera_Pos,
		Object_Pos
	};

public:
	TWorld *p_Load(const char *p_s_file_name);

protected:
	bool Chunk_Read(FILE *fp, long n_pos);
	bool Set_Object_Id(int n_index, int n_id, int n_parent_id);
	int n_Object(const char *p_s_name) const;
	int n_Camera(const char *p_s_name) const;
	int n_Light(const char *p_s_name) const;
	int n_Material(const char *p_s_name) const;
	static bool Read_CStr(char *p_s_name, FILE *fp);
	static bool Read_Skip(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Skip_Float(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Skip_Word(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Skip_DWord(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_Color_RGB(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_Color_RGBF(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_NPercentage(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_FPercentage(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_Light(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_SpotLight(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_LightInnerRange(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_LightOuterRange(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_LightMultiplier(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_LightOff(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_LightShadowed(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MeshName(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MapFile(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_NodeId(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TriObject(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_VertexList(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_VertexFlags(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_FaceList(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_FaceMaterial(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MapList(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_SmoothingList(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MeshMatrix(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_Camera(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MaterialName(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MaterialAmbient(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MaterialDiffuse(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MaterialSpecular(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_MaterialSelfIllum(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_Frames(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackNodeName(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_PivotPoint(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_SplineFlags(FILE *fp, unsigned short &r_n_time,
		float &r_f_tens, float &r_f_cont, float &r_f_bias);
	static bool Read_TrackFlags(FILE *fp, unsigned short &r_n_frame_num);
	static bool Read_TrackPos(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackColor(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackRot(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackFOV(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackRoll(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackScale(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackCamera(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackCamTgt(C3DSLoader *p_this, FILE *fp, long n_pos);
	static bool Read_TrackInfo(C3DSLoader *p_this, FILE *fp, long n_pos);
	int n_Chunk(unsigned short n_id) const;
};

#define LOAD_3DS_H
#endif
